/**
 * UDP client for the contest
 * Copyright 2014, The 2nd ARC/CPSY/RECONF High-Performance Computer System Design Contest
 */


package jp.ac.utsunomiya.is;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;


public class UDP_Client extends Thread {
	protected final int BUF_SIZE = 1024;
	protected byte[] sendBuf = new byte[BUF_SIZE];
	protected byte[] recvBuf = new byte[BUF_SIZE];
	protected boolean verbose = true;
	protected long waitNanoSec = 20000000; // 20 ms

	protected DatagramSocket udpSocket;
	protected DatagramPacket udpPacketSend = null;
	protected DatagramPacket udpPacketRecv = new DatagramPacket(sendBuf, BUF_SIZE);
	protected InetAddress udpDstAddress;
	protected InetAddress udpSrcAddress;
	protected int udpDstPort;
	protected int udpSrcPort;

	protected boolean fileout = false;
	
	protected String LogToUDP = "ToUDP.bin";
	protected String LogFromUDP = "FromUDP.bin";
	private String dataFileName;
	
	public UDP_Client(String targetHost, int targetPort, String filename) {
		this.targetHost = targetHost;
		this.targetPort = targetPort;
		this.dataFileName = filename;
		this.fileout=false;
		this.waitNanoSec = 20000000; // 20 ms

		try {
			udpSocket = new DatagramSocket();
			udpDstAddress = InetAddress.getByName(targetHost);
			udpDstPort = targetPort;
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	String targetHost;
	int targetPort;
	boolean running;
	
	void reset() {
		// RESET packet (size= 16/4 = 4 word)
		try {
			System.out.println("Sending reset packet to " + udpDstAddress +" " + udpDstPort);
			sendBuf = new byte[16];
			DatagramPacket sendInitPacket = new DatagramPacket(sendBuf, 16, udpDstAddress, udpDstPort);
			udpSocket.send(sendInitPacket);
			Thread.sleep(100);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	@Override
	public void run() {
		try {
			udpSocket = new DatagramSocket();
			udpSrcAddress = udpSocket.getLocalAddress();
			udpSrcPort = udpSocket.getLocalPort();

			System.out.println("UDP Socket created. sending file '" 
			+ dataFileName + "' started!");
			Thread f = new Forward();
			Thread b = new Backward();
			
			long before = System.nanoTime();
			f.start();
			b.start();
			
			f.join();
			b.join();
			long after = System.nanoTime();

			System.out.println();
			System.out.println("finished!");
			System.out.println("after-before = " + (after-before) + " (ns) = " 
			+  ((double)(after-before)/1000000000) +"(s)");
			
			} catch(Exception e) {
			e.printStackTrace();
		}
	}
	
	class Forward extends Thread {
		protected FileOutputStream fileToUDP = null;
		@Override
		public void run() {
			int length = 1024;
			FileInputStream fileData = null;
			try {
				//System.out.println("Forward thread started!");
				if (fileout) {
					fileToUDP = new FileOutputStream(LogToUDP);
				}
				
				fileData = new FileInputStream(dataFileName);

				int index=0;
				long prevTime = System.nanoTime();
				while(true) {
					int ret = fileData.read(sendBuf, 0, length);
					if(ret<0) break; // reached to the end of the file
					if(fileout) {
						fileToUDP.write(sendBuf);
					}
					
					DatagramPacket sendPacket = new DatagramPacket(sendBuf, length, udpDstAddress, udpDstPort);

					
					long currentTime = System.nanoTime();
					long elapsedTime = currentTime - prevTime;
					long sleepNanoSec = waitNanoSec - elapsedTime; 
					if(sleepNanoSec < 0) throw new Exception("ERROR : Send packet time overflow!!!");
					int sleepMilliSec = (int) (sleepNanoSec / 1000000);
					//System.out.println("Sleeping (ms) "+sleepMilliSec);
					Thread.sleep(sleepMilliSec);
					//if(index==2) break;
					//prevTime = prevTime + waitNanoSec;
					prevTime = System.nanoTime();
					
					udpSocket.send(sendPacket);
					
					index++;
					//if(index%64==0) System.out.println(index+"KB sent");

				}
			} catch (Exception e) {
				e.printStackTrace();
			} finally {
				//System.out.println("Forward thread finished finally.");
				try {
					fileData.close();
					if (fileout) {
						fileToUDP.close();
					}
				} catch(Exception e) {
					e.printStackTrace();
				}
			}
		}
	}
	
	class Backward extends Thread {
		protected FileOutputStream fileFromUDP = null;
		@Override
		public void run() {
			int length = 0;
			try {
				//System.out.println("Backward thread started!");
				if (fileout) {
					fileFromUDP = new FileOutputStream(LogFromUDP);
				}
				running = true;
				while (running) {
					DatagramPacket recvPacket = new DatagramPacket(recvBuf, BUF_SIZE);
					udpSocket.setSoTimeout(1000*60*10); // wait for 10 min
					//long timeout = udpSocket.getSoTimeout();
					//System.out.println("UDP SoTimeout = " + timeout + " (ms)");
					udpSocket.receive(recvPacket);
					length = recvPacket.getLength();
					if (length<0) {
						throw new IOException("Socket read failed code=" + length);
					}	
					if (length==0) continue;
					
					if(fileout) {
						fileFromUDP.write(recvBuf);
					}
					
					if(true){
						char c0=' ',c1=' ',c2=' ';
						for(int i=0; i<length; i++) {
							char c = (char) recvBuf[i];
							if(verbose) {
								if(c != '\0') System.out.print(c);
							}
	
							c0=c1;
							c1=c2;
							c2=c;
							if(c0 == 'E' && c1 == 'N' && c2 == 'D') {
								running = false;
								break;
							}
						}
					}
					
				}
			} catch (Exception e) {
				e.printStackTrace();
			} finally {
				//System.out.println("Backward thread finished finally.");
				try {
					if (fileout) {
						fileFromUDP.close();
					}
				} catch(Exception e) {
					e.printStackTrace();
				}
			}

		}
	}




	/**
	 * @param args
	 */
	public static void main(String[] args) {
		if(args.length<1 || args.length>2) {
			System.out.println("usage: java jp.ac.utsunomiya.is.UDP_Client targetHost [fileName]");
			System.out.println("example(1): java jp.ac.utsunomiya.is.UDP_Client 192.168.10.64 310sort512.bin");
			System.out.println("example(2): java jp.ac.utsunomiya.is.UDP_Client 192.168.10.64 ");
			System.out.println("(2) for all the application (sort, mm, stencil, spath).");
			System.exit(-1);
		}
		
		String targetHost = args[0];
		int targetPort    = 8100;
		System.out.println("targetHost:"+targetHost+" targetPort:"+targetPort);

		if(args.length==1) 
		{
			new UDP_Client(targetHost, targetPort, "").reset();
			new UDP_Client(targetHost, targetPort, "310sort512.bin").run();
			new UDP_Client(targetHost, targetPort, "").reset();
			new UDP_Client(targetHost, targetPort, "320mm512.bin").run();
			new UDP_Client(targetHost, targetPort, "").reset();
			new UDP_Client(targetHost, targetPort, "330stencil512.bin").run();
			new UDP_Client(targetHost, targetPort, "").reset();
			new UDP_Client(targetHost, targetPort, "340spath512.bin").run();
		} else {
			new UDP_Client(targetHost, targetPort, "").reset();
			new UDP_Client(targetHost, targetPort, args[1]).run();
			//new UDP_Client(targetHost, targetPort, "199transfer512.bin").run();
		}
	}

}
